from scipy.spatial.distance import pdist, squareform
import numpy as np


def distcorr(X, Y):
    """ Compute the distance correlation function.
    Taken from: https://gist.github.com/wladston/c931b1495184fbb99bec

    >>> a = [1,2,3,4,5]
    >>> b = np.array([1,2,9,4,4])
    >>> distcorr(a, b)
    0.762676242417
    """
    X = np.atleast_1d(X)
    Y = np.atleast_1d(Y)
    if np.prod(X.shape) == len(X):
        X = X[:, None]
    if np.prod(Y.shape) == len(Y):
        Y = Y[:, None]
    X = np.atleast_2d(X)
    Y = np.atleast_2d(Y)
    n = X.shape[0]
    if Y.shape[0] != X.shape[0]:
        raise ValueError('Number of samples must match')
    a = squareform(pdist(X))
    b = squareform(pdist(Y))
    A = a - a.mean(axis=0)[None, :] - a.mean(axis=1)[:, None] + a.mean()
    B = b - b.mean(axis=0)[None, :] - b.mean(axis=1)[:, None] + b.mean()

    dcov2_xy = (A * B).sum() / float(n * n)
    dcov2_xx = (A * A).sum() / float(n * n)
    dcov2_yy = (B * B).sum() / float(n * n)
    dcor = np.sqrt(dcov2_xy) / np.sqrt(np.sqrt(dcov2_xx) * np.sqrt(dcov2_yy))
    return dcor



import os
from collections import Counter
from zipfile import ZipFile
from tempfile import TemporaryDirectory

import numpy as np
import pandas as pd


def _process_sources(source_files):
    if not isinstance(source_files, (list, tuple)):
        source_files = [source_files]

    source_files_list = []
    source_files_list.extend(source_files)

    original_source_names_files = {}

    sources = {}

    while len(source_files_list) > 0:
        file_path = source_files_list.pop()

        file_name, file_ext = os.path.splitext(file_path)
        file_ext = file_ext.lower()
        file_name = os.path.basename(file_name)

        if file_ext == '.csv':
            sources[file_name] = pd.read_csv(file_path, index_col=0, mangle_dupe_cols=True)

        elif file_ext in ('.xls', '.xlsx'):
            excel_data = pd.read_excel(file_path, index_col=0, sheet_name=None)
            for sheet_name, sheet_data in excel_data.items():
                sheet_data.columns = sheet_data.columns.astype(str)

                # handle duplicated source names
                original_source_name = sheet_name
                source_name = original_source_name
                if source_name in sources.keys():
                    # rename previously added sources with original name
                    prev_source_data = sources.pop(source_name)
                    prev_source_file_name = original_source_names_files.pop(source_name)
                    prev_source_new_name = '{0} ({1})'.format(source_name, prev_source_file_name)
                    sources[prev_source_new_name] = prev_source_data

                    # new source name for current source data
                    source_name = '{0} ({1})'.format(source_name, file_name)

                sources[source_name] = sheet_data
                original_source_names_files[original_source_name] = file_name

        elif file_ext == '.zip':
            zipf = ZipFile(file_path)
            tempdir = TemporaryDirectory().name
            zipf.extractall(path=tempdir)

            for zip_member in zipf.infolist():
                source_files_list.append(os.path.join(tempdir, zip_member.filename))

    return sources


def get_sources(source_files, rep_merge=np.mean):
    key_func = lambda x: x.split('.')[0]

    sources = _process_sources(source_files)
    sources_names = []
    data_sources = []

    sources_names_sorted = sorted(sources.keys())

    for source_name in sources_names_sorted:
        source_data = sources[source_name]

        # # FIXME columns with '.' should not be allowed, however the code below breaks things
        # if np.array(['.' in col for col in source_data.columns]).sum():
        #     raise ValueError('Columns cannot have a point (.)')

        # FIXME we don't support aggregating values when there is one categorical variable and repeated columns
        numeric_rows = np.array([np.isreal(x) for x in source_data.iloc[:, 0]])

        processed_data = source_data.iloc[numeric_rows].groupby(key_func, axis=1).apply(rep_merge, axis=1)

        # FIXME related to previous comment.
        if processed_data.shape[1] != source_data.shape[1] and not numeric_rows.all():
            raise ValueError('Categorical rows with repeated columns is not supported.')

        if not numeric_rows.all():
            non_numeric_rows_names = source_data.index[~numeric_rows].tolist()
            processed_data = processed_data.append(source_data.loc[non_numeric_rows_names])

        # there can't be duplicated rows per each source
        if processed_data.index.duplicated().any():
            raise ValueError('There can\'t be duplicated rows per each source')

        data_sources.append(processed_data)
        sources_names.append(source_name)

    return data_sources, sources_names


def merge_sources(source_files, rep_merge=np.mean):
    processed_sources, sources_names = get_sources(source_files, rep_merge=rep_merge)
    sources_names = [sn for sn_idx, sn in enumerate(sources_names)
                     for i in range(processed_sources[sn_idx].shape[0])]

    full_sources = pd.concat(processed_sources, sort=True)

    # renamed duplicated
    if not full_sources.index.is_unique:
        old_index = full_sources.index.tolist()
        index_counter = Counter(old_index)
        new_index = []

        for idx, index_value in enumerate(full_sources.index):
            index_value_count = index_counter[index_value]

            if index_value_count == 1:
                new_index.append(index_value)
            else:
                new_index_value = index_value + u' ({0})'.format(sources_names[idx])
                new_index.append(new_index_value)

        full_sources.index = pd.Index(new_index)

    full_sources.index.rename('features', inplace=True)

    return full_sources, list(full_sources.index), sources_names



from multiprocessing import cpu_count

import numpy as np
from scipy import stats
from scipy.signal import argrelextrema
from scipy.spatial.distance import squareform, cdist
from scipy.cluster.hierarchy import fcluster, linkage
import pandas as pd
from joblib import Parallel, delayed

from sklearn.cluster import KMeans
from sklearn.cluster import SpectralClustering
from sklearn.metrics import adjusted_rand_score as ari
from sklearn.neighbors import KernelDensity
from sklearn.preprocessing import scale


def _get_perc_from_k(k):
    return [(1.0 / k) * i for i in range(1, k)]


def run_quantile_clustering(data, k, **kwargs):
    data_perc = stats.rankdata(data, 'average') / len(data)
    data_perc_sort_idx = np.argsort(data_perc)

    # data_perc = data
    percentiles = [0.0] + _get_perc_from_k(k) + [1.0]

    cut_points = np.searchsorted(data_perc[data_perc_sort_idx], percentiles, side='right')

    current_cluster = 0
    part = np.zeros(data.shape, dtype=float) - 1

    for i in range(len(cut_points) - 1):
        lim1 = cut_points[i]
        lim2 = cut_points[i+1]

        part[data_perc_sort_idx[lim1:lim2]] = current_cluster
        current_cluster += 1

    return part


def categorical_clustering(data, **kwargs):
    cat_clus = {v: k for k, v in enumerate(np.unique(data))}

    return np.array([cat_clus[x] for x in data])


def run_maxdiff_clustering(data, k, **kwargs):
    data_idx_sorted = np.argsort(data)
    data_sorted = data[data_idx_sorted]

    data_diffs = np.array([abs(data_sorted[i] - data_sorted[i+1]) for i in range(len(data_sorted)-1)])
    data_diffs_idx_sort = np.argsort(data_diffs)[::-1]

    break_points = data_diffs_idx_sort[:(k-1)]
    break_points = np.sort(break_points)
    break_points = break_points[::-1]

    partition = np.zeros(len(data), int)
    current_cluster = 1

    for idx in range(len(break_points)):
        curr_bp_idx = break_points[idx] + 1

        partition[data_idx_sorted[:curr_bp_idx]] = current_cluster
        current_cluster += 1

    return partition


def run_kde_clustering(data, k, bandw=1.0, **kwargs):
    # FIXME: the next line was commented out in patch
    data = scale(data)
    data_res = data.reshape((-1,))

    mi = []
    mi_counter = 0

    while len(mi) + 1 < k and mi_counter < 3:
        bandw = bandw / 1.8
        s, kde_sample = _run_kde(data, bandw)

        mi = argrelextrema(kde_sample, np.less)[0]

        mi_counter += 1

    if mi_counter >= 3:
        raise Exception('bandwidth did not produced at least {0} clusters'.format(k))

    samples_idx_sorted = np.argsort(kde_sample[mi])
    samples_idx_sorted = mi[samples_idx_sorted]

    partition = np.zeros(len(data), int)

    current_cluster = 1
    for idx in range(k - 1):
        mi_idx = samples_idx_sorted[idx]
        mi_idx_value = s[mi_idx]

        if idx == 0:
            partition[data_res < mi_idx_value] = current_cluster
            current_cluster += 1

        if idx != (k - 1 - 1):
            mi_next_idx = samples_idx_sorted[idx + 1]
            mi_next_idx_value = s[mi_next_idx]

            if mi_idx_value < mi_next_idx_value:
                partition[(data_res >= mi_idx_value) * (data_res < mi_next_idx_value)] = current_cluster
            else:
                partition[(data_res >= mi_next_idx_value) * (data_res < mi_idx_value)] = current_cluster

            current_cluster += 1
        # else:
        #     partition[(data_res >= mi_idx_value)] = current_cluster

    return partition


def _run_kde(data, bandwidth):
    data_std = data.std()

    kde = KernelDensity(kernel='gaussian', bandwidth=bandwidth).fit(data.reshape(-1, 1))
    s = np.linspace(data.min() - data_std, data.max() + data_std)
    e = kde.score_samples(s.reshape(-1, 1))

    return s, e


def get_squareform(similarity_pdist, fill_diag_value=1.0):
    assert similarity_pdist is not None, 'similarity_pdist is None'
    assert hasattr(similarity_pdist, 'shape'), 'similarity_pdist has no shape attribute'
    assert len(similarity_pdist.shape) == 1, 'similarity_pdist has incorrect shape'

    similarity_square = squareform(similarity_pdist.astype(float))
    np.fill_diagonal(similarity_square, fill_diag_value)

    return similarity_square


def _isempty(row):
    return np.array([x is None or (np.isreal(x) and np.isnan(x)) for x in row])


def _get_common_features(obj1, obj2):
    obj1_notnan = np.logical_not(_isempty(obj1))
    obj2_notnan = np.logical_not(_isempty(obj2))

    common_features = np.logical_and(obj1_notnan, obj2_notnan)
    n_common_features = common_features.sum()

    return common_features, n_common_features


def _get_range_n_clusters(n_common_features, **kwargs):
    internal_n_clusters = kwargs.get('internal_n_clusters')

    if internal_n_clusters is None:
        estimated_k = int(np.floor(np.sqrt(n_common_features)))
        estimated_k = np.min((estimated_k, 10))
        range_n_clusters = range(2, np.max((estimated_k, 3)))
    elif isinstance(internal_n_clusters, (tuple, list, range)):
        range_n_clusters = internal_n_clusters
    elif isinstance(internal_n_clusters, int):
        range_n_clusters = (internal_n_clusters,)
    else:
        raise ValueError('n_clusters is invalid')

    return range_n_clusters


def _get_internal_parts(data_obj, range_n_clusters, clustering_method, **kwargs):
    partitions = []

    for k in range_n_clusters:
        if len(data_obj) <= k:
            part = np.array([np.nan] * len(data_obj))
        else:
            if all([np.isreal(x) for x in data_obj]):
                part = clustering_method(data_obj, k, **kwargs)
                # part_k = len(np.unique(part))
                # assert part_k == k, 'partition has the wrong number of clusters: {}'.format(part_k)
            elif all([not np.isreal(x) for x in data_obj]):
                part = categorical_clustering(data_obj, **kwargs)
            else:
                raise ValueError('Data is not all numerical or not all categorical')

        partitions.append(part)

    return np.array(partitions)


def _get_clustering_method(**kwargs):
    internal_clustering_method = kwargs.get('internal_clustering_method', 'quantiles')

    if internal_clustering_method == 'kmeans':
        return run_kmeans
    elif internal_clustering_method == 'quantiles':
        return run_quantile_clustering
    else:
        raise ValueError('invalid internal clustering method')


def row_col_from_condensed_index(d,i):
    b = 1 -2*d
    x = np.floor((-b - np.sqrt(b**2 - 8*i))/2)
    y = i + x*(b + x + 2)/2 + 1
    return (int(x), int(y))


def _compute_ari(part1, part2):
    if np.isnan(part1).any() or len(part1) == 0:
        return 0.0

    return ari(part1, part2)


def _get_contingency_table(p1, p2):
    df = pd.DataFrame({'p1': p1, 'p2': p2})
    return pd.crosstab(df.p1, df.p2).values


def _get_fisher_pvalue(p1, p2):
    raise NotImplemented()
    # contingency_table = _get_contingency_table(p1, p2)
    #return fisher_exact(contingency_table)


def _get_chi2(p1, p2):
    # n_objects = len(p1)

    p1_clusters = np.unique(p1)
    p2_clusters = np.unique(p2)

    if len(p1_clusters) == 1 or len(p2_clusters) == 1:
        print('Partition with one cluster. Skipping')
        return np.nan, 1.0

    contingency_table = _get_contingency_table(p1, p2)

    chi2, chi2_pvalue, dof, expected = stats.chi2_contingency(contingency_table)

    # k, l = len(p1_clusters), len(p2_clusters)
    # vmes = np.sqrt(chi2 / (n_objects * np.min((k - 1, l - 1))))

    return chi2_pvalue


def _get_perm_set(obj1, obj2, clustering_method, perm_range):
    rs = np.random.RandomState(perm_range[0])
    # print(perm_range[0])

    other_cms = []
    for i in perm_range:
        obj2_permuted = rs.permutation(obj2)

        new_cm = cm(obj1, obj2_permuted, clustering_method=clustering_method)[0]
        other_cms.append(new_cm)

    return other_cms


def _get_perm_pvalue(orig_cm, obj1, obj2, clustering_method, n_perm, perm_n_jobs=1, **kwargs):
    n_cpus = perm_n_jobs if perm_n_jobs > 0 else cpu_count()
    step = int(np.ceil(n_perm / n_cpus))
    full_perm_range = range(0, n_perm, step)

    other_cms = Parallel(n_jobs=perm_n_jobs)(
        delayed(_get_perm_set)(obj1, obj2, clustering_method, perm_range)
        for perm_range in [range(s, min(s + full_perm_range.step, full_perm_range.stop)) for s in full_perm_range]
    )

    # flatten
    other_cms = np.array([cm_val for sublist in other_cms for cm_val in sublist])

    return (len(other_cms[other_cms >= orig_cm]) + 1) / (len(other_cms) + 1)


def cm(obj1, obj2, clustering_method=run_quantile_clustering, compute_perm_pvalue=False,
       compute_chi2_pvalue=False, compute_fisher_pvalue=False, **kwargs):
    range_n_clusters = _get_range_n_clusters(len(obj1), **kwargs)

    obj1_parts = _get_internal_parts(obj1, range_n_clusters, clustering_method, **kwargs)
    obj2_parts = _get_internal_parts(obj2, range_n_clusters, clustering_method, **kwargs)

    comp_values = cdist(obj1_parts, obj2_parts, metric=_compute_ari)

    max_pos = np.unravel_index(comp_values.argmax(), comp_values.shape)
    max_ari = comp_values[max_pos]

    obj1_max_part = obj1_parts[max_pos[0]]
    obj2_max_part = obj2_parts[max_pos[1]]

    if len(np.unique(obj1_max_part)) == 1 or len(np.unique(obj2_max_part)) == 1:
        return 0.0, 1.0

    sim_value = max_ari
    pvalue = None if not compute_fisher_pvalue else _get_fisher_pvalue(obj1_max_part, obj2_max_part)

    if compute_chi2_pvalue and not compute_fisher_pvalue:
        pvalue = _get_chi2(obj1_max_part, obj2_max_part)

    if compute_perm_pvalue:
        pvalue = _get_perm_pvalue(sim_value, obj1, obj2, clustering_method, **kwargs)

    return sim_value, pvalue


def get_shared_objects(obj1, obj2, **kwargs):
    n_common_features = _get_common_features(obj1, obj2)[1]
    return (n_common_features,)


def get_pearson(obj1, obj2, **kwargs):
    if len(np.unique(obj1)) == 1 or len(np.unique(obj2)) == 1 or \
            any([not np.isreal(x) for x in obj1]) or \
            any([not np.isreal(x) for x in obj2]):
        return 0.0, 1.0

    return stats.pearsonr(obj1, obj2)


def _calculate_sub_simmatrix(data, idx_range, sim_func='cm', return_pvalue=False, min_n_common_features=3, **kwargs):
    p_dist = []
    p_dist_pvalue = []
    n_objects = data.shape[0]

    if sim_func == 'cm':
        similarity_func = cm
    elif sim_func == 'pearson':
        similarity_func = get_pearson
    elif sim_func == 'shared_objects':
        similarity_func = get_shared_objects
    else:
        raise ValueError('Invalid sim_func')

    for idx in idx_range:
        obj1_idx, obj2_idx = row_col_from_condensed_index(n_objects, idx)

        obj1 = data[obj1_idx]
        obj2 = data[obj2_idx]

        common_features, n_common_features = _get_common_features(obj1, obj2)

        if n_common_features < min_n_common_features:
            sim_values = (0.0, 1.0) # sim value and pvalue
        else:
            sim_values = similarity_func(obj1[common_features], obj2[common_features], **kwargs)

        p_dist.append(sim_values[0])

        if return_pvalue:
            p_dist_pvalue.append(sim_values[1])

    return p_dist, p_dist_pvalue


def calculate_simmatrix(data, fill_diag_value=1.0, n_jobs=1, **kwargs):
    data_index = None
    if hasattr(data, 'index'):
        data_index = data.index.tolist()

    if hasattr(data, 'values'):
        data = data.values

    # FIXME: quantiles clustering is only for 1d comparisons. Use kmeans for n-dimensional.
    kwargs['clustering_method'] = _get_clustering_method(**kwargs)

    n_objects = data.shape[0]

    p_dist_len = int((n_objects * (n_objects - 1)) / 2)

    # FIXME: set n_jobs according to data size. Do some performance test with unit tests
    n_cpus = n_jobs if n_jobs > 0 else cpu_count()

    step = int(np.ceil(p_dist_len / n_cpus))
    p_dist_range = range(0, p_dist_len, step)

    p_dist_values = Parallel(n_jobs=n_jobs)(
        delayed(_calculate_sub_simmatrix)(data, idx_range, **kwargs)
        for idx_range in [range(s, min(s + p_dist_range.step, p_dist_range.stop)) for s in p_dist_range]
    )

    p_dist = []
    p_dist_pval = []
    for p, pval in p_dist_values:
        p_dist.extend(p)

        if len(pval) > 0:
            p_dist_pval.extend(pval)

    p_dist = np.array(p_dist)
    p_dist_pval = np.array(p_dist_pval)

    return_pvalue = kwargs.get('return_pvalue', False)

    if data_index is not None:
        sqmatrix = get_squareform(p_dist, fill_diag_value)
        sim_matrix = pd.DataFrame(
            sqmatrix,
            index=data_index,
            columns=data_index
        )

        if return_pvalue:
            sqmatrix_pval = get_squareform(p_dist_pval, np.nan)
            pval_matrix = pd.DataFrame(
                sqmatrix_pval,
                index=data_index,
                columns=data_index
            )
    else:
        sim_matrix = p_dist
        pval_matrix = p_dist_pval

    if return_pvalue:
        return sim_matrix, pval_matrix
    else:
        return sim_matrix


def _get_partition(clustering_method, n_clusters, sim_matrix):
    # get all partitions
    partitions = {}
    for k in n_clusters:
        column_name = 'k={0}'.format(str(k))
        partitions[column_name] = clustering_method(k)

    final_partition = pd.DataFrame(partitions)

    if hasattr(sim_matrix, 'index'):
        final_partition = final_partition.set_index(sim_matrix.index)

    if len(n_clusters) == 1 and not hasattr(sim_matrix, 'index'):
        final_partition = final_partition.values.reshape(-1)

    return final_partition


def get_normalized_sim_matrix(sim_matrix):
    sim_matrix_min = sim_matrix.min()
    sim_matrix_max = sim_matrix.max()

    if len(sim_matrix.shape) == 2:
        sim_matrix_min = sim_matrix_min.min()
        sim_matrix_max = sim_matrix_max.max()

    return ((sim_matrix - sim_matrix_min) / (sim_matrix_max - sim_matrix_min))


def get_sim_matrix_by_partition(sim_matrix, partition):
    # pick first partition
    part = partition.iloc[:, 0]

    sorted_index = []

    new_matrix_vals = np.empty(sim_matrix.shape)
    new_matrix_vals[:] = np.nan
    pval_matrix_sorted = pd.DataFrame(
        new_matrix_vals,
        index=sim_matrix.index.tolist(),
        columns=sim_matrix.index.tolist()
    )

    for k in np.unique(part):
        cluster_index = part[part == k].index.tolist()
        sorted_index.extend(cluster_index)

        pval_matrix_sorted.loc[cluster_index, cluster_index] = sim_matrix.loc[cluster_index, cluster_index]

    return pval_matrix_sorted.loc[sorted_index, sorted_index]


def get_pval_matrix_by_partition(data, partition, k_internal, min_n_common_features, n_perms, n_jobs):
    # pick first partition
    part = partition.iloc[:, 0]

    sorted_index = []

    new_matrix_vals = np.empty((data.shape[0], data.shape[0]))
    new_matrix_vals[:] = np.nan
    pval_matrix_sorted = pd.DataFrame(
        new_matrix_vals,
        index=data.index.tolist(),
        columns=data.index.tolist()
    )

    for k in np.unique(part):
        cluster_index = part[part == k].index.tolist()

        cluster_k_pvalue_matrix = \
            calculate_simmatrix(data.loc[cluster_index], internal_n_clusters=k_internal, compute_perm_pvalue=True,
                                n_perm=n_perms, return_pvalue=True, min_n_common_features=min_n_common_features,
                                n_jobs=n_jobs)[1]

        sorted_index.extend(cluster_index)

        pval_matrix_sorted.loc[cluster_index, cluster_index] = cluster_k_pvalue_matrix.loc[cluster_index, cluster_index]

    return pval_matrix_sorted.loc[sorted_index, sorted_index]


def get_partition_spectral(sim_matrix, n_clusters, **kwargs):
    if len(sim_matrix.shape) == 1:
        sim_matrix = get_squareform(sim_matrix)

    if not hasattr(n_clusters, '__iter__'):
        n_clusters = (n_clusters,)

    norm_sim_matrix = get_normalized_sim_matrix(sim_matrix)

    def clustering_method(k):
        return SpectralClustering(n_clusters=k, affinity='precomputed', **kwargs).fit_predict(norm_sim_matrix)

    return _get_partition(clustering_method, n_clusters, sim_matrix)


def get_partition_agglomerative(sim_matrix, n_clusters, linkage_method='average', criterion='maxclust'):
    if len(sim_matrix.shape) == 1:
        sim_matrix = get_squareform(sim_matrix)

    if not hasattr(n_clusters, '__iter__'):
        n_clusters = (n_clusters,)

    # condensed disimilarity matrix for hierarchical methods
    norm_sim_matrix = get_normalized_sim_matrix(sim_matrix)
    norm_disim_pdist = 1 - squareform(norm_sim_matrix, checks=False)

    # FIXME: maybe we should warn if method is centroid, median or ward, as distances are not euclidean
    linkage_matrix = linkage(norm_disim_pdist, method=linkage_method, metric=None)

    def clustering_method(k):
        return fcluster(linkage_matrix, k, criterion=criterion)

    return _get_partition(clustering_method, n_clusters, sim_matrix), linkage_matrix


def clustermatch(data, n_clusters, **kargs):
    sim_matrix = calculate_simmatrix(data, **kargs)

    return get_partition_spectral(sim_matrix, n_clusters), get_normalized_sim_matrix(sim_matrix)


def run_kmeans(data, k, **kwargs):
    kmeans_n_init = 3
    if 'kmeans_n_init' in kwargs:
        kmeans_n_init = kwargs['kmeans_n_init']

    kmeans_random_state = None
    if 'kmeans_random_state' in kwargs:
        kmeans_random_state = kwargs['kmeans_random_state']

    kmeans = KMeans(n_clusters=k, init='k-means++', n_init=kmeans_n_init, random_state=kmeans_random_state)

    return kmeans.fit_predict(data.reshape(-1, 1))



import tempfile


def get_temp_file_name(file_extension=''):
    if not file_extension.startswith('.'):
        file_extension = '.' + file_extension

    with tempfile.NamedTemporaryFile(suffix=file_extension, delete=False) as tmpfile:
        temp_file_name = tmpfile.name
    return temp_file_name



import os
import time
from os.path import dirname, join
import shutil
import json

import numpy as np
from scipy.cluster.hierarchy import fcluster
from scipy.spatial.distance import squareform, pdist
import pandas as pd

try:
    import matplotlib
    import seaborn as sns
    sns.set(context="paper", font="monospace")
    MATPLOTLIB_INSTALLED = True
except:
    MATPLOTLIB_INSTALLED = False

try:
    import requests
    REQUESTS_INSTALLED = True
except:
    REQUESTS_INSTALLED = False

from utils.misc import get_temp_file_name


RESULTS_DIR = 'results'


def _get_condensed_distance_matrix(ensemble):
    return pdist(ensemble.T, lambda u, v: (u != v).sum() / len(u))


def get_timestamp():
    return time.strftime('%Y%m%d_%H%M%S')


def setup_results_dir(func):
    def func_wrapper(*args, **kwargs):
        results_dir = os.path.join(RESULTS_DIR, kwargs['timestamp'])
        if not os.path.isdir(results_dir):
            os.makedirs(results_dir)
        return func(*args, **kwargs)

    return func_wrapper


def get_clustergrammer_link(square_matrix, names):
    if not REQUESTS_INSTALLED:
        raise ValueError('requests is not installed')

    # save sim matrix as csv and get clustergrammer visualization
    df = pd.DataFrame(square_matrix)
    df['names'] = names
    df = df.set_index('names')
    df.columns = names
    square_matrix_file = get_temp_file_name('txt')
    df.to_csv(square_matrix_file, sep='\t', encoding='utf-8')

    clustergrammer_link = ''

    try:
        upload_url = 'http://amp.pharm.mssm.edu/clustergrammer/matrix_upload/'
        r = requests.post(upload_url, files={'file': open(square_matrix_file, 'rb')})
        clustergrammer_link = r.text
    except:
        pass

    return clustergrammer_link


@setup_results_dir
def to_binary(data, file_name, timestamp):
    file_path = os.path.join(RESULTS_DIR, timestamp, file_name + '.pkl')
    pd.to_pickle(data, file_path)


def from_binary(file_name):
    return pd.read_pickle(file_name)


@setup_results_dir
def create_partition_plot_html(partition, timestamp, sources=None):
    results_dir = os.path.join(RESULTS_DIR, timestamp)

    html_dir = join(dirname(__file__), 'html')

    for afile in os.listdir(html_dir):
        afile_path = join(html_dir, afile)
        shutil.copy(afile_path, results_dir)

    if sources is not None:
        sources = np.array(sources)
    else:
        sources = np.array([''] * len(partition))

    # FIXME: create json and copy
    json_path = os.path.join(RESULTS_DIR, timestamp, 'data' + '.json')

    cluster_children = []
    k_values = np.unique(partition)
    for cluster_number in k_values:
        idx = (partition == cluster_number).values
        cluster_objects = partition[idx].index.tolist()
        cluster_objects_sources = sources[idx]

        cluster_children.append(
            {'name': '',
             'children': [{'name': obj_name, 'source': obj_source, 'size': 1}
                          for obj_name, obj_source in zip(cluster_objects, cluster_objects_sources)]
             }
        )

    partition_json = {'name': '', 'children': cluster_children}

    with open(json_path, 'w') as jf:
        json.dump(partition_json, jf)

    return join(results_dir, 'index.html')


@setup_results_dir
def save_ensembles(ensembles, timestamp):
    for ens_name, ens in ensembles.items():
        ensemble_path = os.path.join(RESULTS_DIR, timestamp, ens_name + '_ensemble' + '.xls')
        ens.to_excel(ensemble_path, encoding='utf-8')


@setup_results_dir
def append_data_description(text, timestamp):
    filepath = os.path.join(RESULTS_DIR, timestamp, 'README.txt')
    with open(filepath, 'a') as f:
        f.write('\n')
        f.write(text + '\n')


@setup_results_dir
def write_text_file(text_content, file_name, timestamp):
    filepath = os.path.join(RESULTS_DIR, timestamp, file_name)
    with open(filepath, 'w') as f:
        f.write(text_content)


@setup_results_dir
def write_data_description(data_files, merged_sources, feature_names, sources_names, timestamp):
    filepath = os.path.join(RESULTS_DIR, timestamp, 'README.txt')
    with open(filepath, 'w') as f:
        f.write('Data files included:\n')
        for afile in data_files:
            f.write('  - {}\n'.format(afile))

        f.write('\n')
        f.write('Merged sources shape: {}\n'.format(merged_sources.shape))
        f.write('Number of features: {} ({} unique)\n'.format(len(feature_names), len(set(feature_names))))
        f.write('Number of sources: {}\n'.format(len(set(sources_names))))


@setup_results_dir
def save_excel(dataframe, filename, timestamp):
    filepath = os.path.join(RESULTS_DIR, timestamp, filename + '.xlsx')
    dataframe.to_excel(filepath, encoding='utf-8')


def save_partitions_simple(partitions, partitions_path, extra_columns=None, columns_order=None, sort_by_columns=None):
    if extra_columns is not None:
        extra_df = pd.DataFrame(extra_columns, index=partitions.index)
        partitions = pd.concat([partitions, extra_df], axis=1)

    if columns_order is not None:
        partitions = partitions[columns_order]

    if sort_by_columns is not None:
        partitions = partitions.sort_values(sort_by_columns)

    partitions.to_excel(partitions_path, encoding='utf-8')


@setup_results_dir
def save_partitions(partitions, timestamp, **kwargs):
    partitions_path = os.path.join(RESULTS_DIR, timestamp, 'partitions' + '.xls')
    save_partitions_simple(partitions, partitions_path, **kwargs)

    return partitions_path


@setup_results_dir
def save_coassociation_matrix(ensemble, partition, timestamp, columns_order=None, image_format='pdf'):
    if not MATPLOTLIB_INSTALLED:
        raise ValueError('matplotlib is not installed')

    condensed_matrix = _get_condensed_distance_matrix(ensemble)
    full_matrix = squareform(condensed_matrix)
    full_matrix = 1 - full_matrix
    np.fill_diagonal(full_matrix, 0.0)

    tomato_accessions = ensemble.columns

    if columns_order is None:
        ordering_partition = partition.columns[0]
    else:
        ordering_partition = columns_order[0]

    partition_idxs_sorted = np.argsort(partition[ordering_partition])
    sim_matrix_sorted = full_matrix[partition_idxs_sorted, :]
    sim_matrix_sorted = sim_matrix_sorted[:, partition_idxs_sorted]
    tomato_accessions_sorted = [tomato_accessions[idx] for idx in partition_idxs_sorted]

    df = pd.DataFrame(sim_matrix_sorted, index=tomato_accessions_sorted, columns=tomato_accessions_sorted)
    # fig = sns.plt.figure(figsize=(100, 100))
    ax = sns.heatmap(df, square=True, linewidths=.05, cmap='Blues', cbar=None)
    sns.plt.yticks(rotation=0)
    sns.plt.xticks(rotation=90)

    font_scale = (35.0 / len(tomato_accessions))
    for label in (ax.get_xticklabels() + ax.get_yticklabels()):
        label.set_fontsize(label.get_fontsize() * font_scale)

    sim_matrix_path = os.path.join(RESULTS_DIR, timestamp, 'coasociation_matrix.' + image_format)
    sns.plt.savefig(sim_matrix_path, dpi=300, bbox_inches='tight')
    sns.plt.close()

    sim_matrix_csv_path = os.path.join(RESULTS_DIR, timestamp, 'coasociation_matrix.csv')
    df.to_csv(sim_matrix_csv_path)

    return sim_matrix_path, full_matrix, sim_matrix_csv_path


@setup_results_dir
def save_similarity_matrix(partition, feature_names, sim_matrix, timestamp, image_format='pdf'):
    if not MATPLOTLIB_INSTALLED:
        raise ValueError('matplotlib is not installed')

    partition_idxs_sorted = np.argsort(partition)
    sim_matrix_sorted = sim_matrix[partition_idxs_sorted, :]
    sim_matrix_sorted = sim_matrix_sorted[:, partition_idxs_sorted]
    feature_names_sorted = [feature_names[idx] for idx in partition_idxs_sorted]

    df = pd.DataFrame(sim_matrix_sorted, index=feature_names_sorted, columns=feature_names_sorted)
    ax = sns.heatmap(df, square=True, linewidths=.05)

    sns.plt.yticks(rotation=0)
    sns.plt.xticks(rotation=90)

    for label in (ax.get_xticklabels() + ax.get_yticklabels()):
        label.set_fontsize(1.5)

    sim_matrix_path = os.path.join(RESULTS_DIR, timestamp, 'similarity_matrix.' + image_format)
    sns.plt.savefig(sim_matrix_path, dpi=300, bbox_inches='tight')

    return sim_matrix_path


@setup_results_dir
def save_reps_comparison(reps_comparison, timestamp):
    reps_comparison_csv = os.path.join(RESULTS_DIR, timestamp, 'reps_comparison.csv')
    reps_comparison.to_csv(reps_comparison_csv)
    return reps_comparison_csv


@setup_results_dir
def save_clustermap(sim_matrix, feature_names, sources_names, partition_linkage, timestamp, image_format='pdf'):
    if not MATPLOTLIB_INSTALLED:
        raise ValueError('matplotlib is not installed')

    font_scale = (50.0 / len(feature_names))
    # sns.set(font_scale=font_scale)

    fig = sns.plt.figure(figsize=(100, 100))

    df = pd.DataFrame(sim_matrix, index=feature_names, columns=feature_names)

    part00_k = 4
    part00 = fcluster(partition_linkage, part00_k, criterion='maxclust') - 1

    part01_k = 10
    part01 = fcluster(partition_linkage, part01_k, criterion='maxclust') - 1

    part02_k = 15
    part02 = fcluster(partition_linkage, part02_k, criterion='maxclust') - 1

    part03_k = 20
    part03 = fcluster(partition_linkage, part03_k, criterion='maxclust') - 1

    part00_colors = sns.color_palette('pastel', len(np.unique(part00)))
    part01_colors = sns.color_palette('pastel', len(np.unique(part01)))
    part02_colors = sns.color_palette('pastel', len(np.unique(part02)))
    part03_colors = sns.color_palette('pastel', len(np.unique(part03)))

    df_colors = pd.DataFrame(index=df.index)
    df_colors['$k$={0}'.format(part00_k)] = [part00_colors[i] for i in part00]
    df_colors['$k$={0}'.format(part01_k)] = [part01_colors[i] for i in part01]
    df_colors['$k$={0}'.format(part02_k)] = [part02_colors[i] for i in part02]
    df_colors['$k$={0}'.format(part03_k)] = [part03_colors[i] for i in part03]

    cm = sns.clustermap(df, col_linkage=partition_linkage, row_linkage=partition_linkage,
                        row_colors=df_colors, col_colors=df_colors)
    sns.plt.setp(cm.ax_heatmap.xaxis.get_majorticklabels(), rotation=90)
    sns.plt.setp(cm.ax_heatmap.yaxis.get_majorticklabels(), rotation=0)

    ax = cm.cax
    pos1 = ax.get_position()  # get the original position
    pos2 = [pos1.x0 + 0.79, pos1.y0 - 0.02, pos1.width, pos1.height]
    ax.set_position(pos2)  # set a new position

    unique_sources = list(set(sources_names))
    sources_colors = sns.color_palette('hls', len(unique_sources))
    ylabels = [x for x in cm.ax_heatmap.get_yticklabels()]
    ylabels_text = [x.get_text() for x in ylabels]

    leg_fig = cm.ax_heatmap.get_yticklabels()[0].figure

    for label in cm.ax_heatmap.get_xticklabels():
        label_idx = feature_names.index(label.get_text())
        label_source = sources_names[label_idx]
        color_idx = unique_sources.index(label_source)
        label.set_color(sources_colors[color_idx])
        label.set_fontsize(label.get_fontsize() * font_scale)

        ylabel = ylabels[ylabels_text.index(label.get_text())]
        ylabel.set_color(sources_colors[color_idx])
        ylabel.set_fontsize(ylabel.get_fontsize() * font_scale)

    legend_labels = unique_sources
    legend_patches = [matplotlib.patches.Patch(color=C, label=L) for
                      C, L in zip(sources_colors, legend_labels)]
    leg_fig.legend(handles=legend_patches, labels=legend_labels, loc='lower right', bbox_to_anchor=(0.970, 0.05))

    clustermap_path = os.path.join(RESULTS_DIR, timestamp, 'clustermap.' + image_format)
    cm.savefig(clustermap_path, dpi=300, bbox_inches='tight')
    sns.plt.close()

    clustermap_csv_path = os.path.join(RESULTS_DIR, timestamp, 'clustermap.' + 'csv')
    df.to_csv(clustermap_csv_path)

    return clustermap_path, clustermap_csv_path




import argparse
import logging

from clustermatch import __short_description__
from clustermatch.cluster import calculate_simmatrix, get_partition_spectral, get_pval_matrix_by_partition
from utils.data import merge_sources
from utils.output import save_partitions_simple

LOG_FORMAT = "[%(asctime)s] %(levelname)s: %(message)s"
logging.basicConfig(format=LOG_FORMAT, level=logging.INFO)
logger = logging.getLogger('root')


def run():
    parser = argparse.ArgumentParser(description=__short_description__)

    # Mandatory parameters
    parser.add_argument('-i', '--input-files', required=True, type=str, nargs='+', help=
        'Path to input data files (could be one or multiple files). It could be a csv, xls (with different worksheets) or zip file.'
    )
    parser.add_argument('-k', '--n-clusters', required=True, type=int, nargs='+', help=
        'Number of final clusters (could contain multiple values).'
    )
    parser.add_argument('-o', '--output-file', required=True, type=str, help=
        'Path to output data partition file. The extension'
    )

    # Optional parameters
    parser.add_argument('--n-init', type=int, default=10, help=
        'Number of time the k-means algorithm will be run with different centroid seeds. '
        'The final results will be the best output of n_init consecutive runs in terms of inertia.'
    )
    parser.add_argument('--n-jobs', type=int, default=1, help=
        'The number of parallel jobs to run. -1 means using all processors.'
    )
    parser.add_argument('--minimum-objects', type=int, default=5, help=
        'Minimum amount of objects shared between two features to process them.'
    )

    args = parser.parse_args()

    # Validate parameters
    if any(x < 2 for x in args.n_clusters):
        parser.error('Number of final clusters must be >= 2')

    # Read data files
    logger.info('Reading input data files')
    merged_sources, feature_names, sources_names = merge_sources(args.input_files)

    # Run clustermatch
    logger.info(f'Getting similarity matrix for {merged_sources.shape[0]} variables')
    cm_sim_matrix = calculate_simmatrix(merged_sources, min_n_common_features=args.minimum_objects, n_jobs=args.n_jobs)

    logger.info(f'Running spectral clustering with k={args.n_clusters}')
    partition = get_partition_spectral(cm_sim_matrix, args.n_clusters, n_init=args.n_init, n_jobs=args.n_jobs)

    # if args.compute_pvalues:
    #     print('Getting pvalue matrix')
    #     cm_pvalue_sim_matrix = get_pval_matrix_by_partition(
    #         merged_sources, partition,
    #         k_internal, min_n_tomatoes,
    #         args.compute_pvalues_n_perms,
    #         n_jobs
    #     )
    #
    #     save_excel(cm_pvalue_sim_matrix, 'cm_pvalue', timestamp=timestamp)
    #     print('cm_pvalue saved')

    columns_order = ['k={0}'.format(str(k)) for k in args.n_clusters]

    logger.info(f'Saving partition to {args.output_file}')
    save_partitions_simple(
        partition,
        args.output_file,
        extra_columns={'sources': sources_names},
        columns_order=['sources', *columns_order],
        sort_by_columns=columns_order
    )



import setuptools
from clustermatch import __version__, __short_description__

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="clustermatch",
    version=__version__,
    author="Milton Pividori",
    author_email="miltondp@uchicago.edu",
    description=__short_description__,
    license='MIT',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/sinc-lab/clustermatch",
    packages=['clustermatch', 'utils'],
    python_requires='>=3',
    install_requires=[
        'numpy',
        'scipy',
        'pandas',
        'joblib',
        'scikit-learn',
        'xlrd',
        'xlwt',
        'openpyxl',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console"
    ],
    entry_points={
        'console_scripts': [
            'clustermatch = clustermatch.main:run'
        ]
    },
)
